program tbcnrand

!   This tests the sequential version of the BCN pseudorandom number generator.
!   David H Bailey    2008-08-25

!   The sequence generated is (x_k), where x_k = first 53 binary digits of the
!   binary expansion of alpha_{2,3} beginning at binary position 53*k + aa,
!   normalized by 2^53.  The parameter aa is the starting point of the sequence,
!   or in other words the seed of the pseudorandom sequence.  To obtain the 
!   maximum period, aa should be set to at least 3^33 = 5.5590606e15.  The value
!   of aa must not exceed 2^53 = 9.0071992e15.  When aa is set in this range, the
!   period of the generator is 2x3^32 = 3.7060404e15.  In general, the period
!   of the generator is 2x3^(p-1), where 3^p is the largest power of 3 less 
!   than aa.  However the generated sequence will not match the binary digits of
!   alpha_{2,3} if the range of sequence indices, namely [aa, aa+53*n], includes
!   a power of 3.

!   The bcnrand routine is designed for simple parallelization.  For example,
!   in an MPI program, suppose that kpr is the processor number and npr is the
!   number of processors.  Then the line 
!       call bcnrand (n/npr, aa + 53*n/npr*kpr, x)
!   generates on each processor a section of length n/npr.  In this way, the
!   npr processors collectively have the same n-long sequence (provided that
!   n is divisible by npr) as is generated on a single processor system by
!   means of the line
!       call bcnrand (n, aa, x)

!   On certain systems with a fused multiply-add instruction, see !> comments.

!   Parameters:
!     n    Length of the sequence desired.
!     np   Number of initial values that are output at start and end.
!     aa   Seed of sequence; default = 3^33 + 100; should satisfy
!          3^33 + 100 <= aa <= 2^53.

!   NOTE:  This includes a dummy "second" timing function below.  Replace
!   with a system-specific verison for nonzero output run times.

implicit none
integer i, n, np
parameter (n = 100000000, np = 10)
real*8 aa, x(n), tm0, tm1
parameter (aa = 3.d0**33 + 100.d0)

write (6, 1) n, aa
1 format ('bcnrand test program'/'n =',i10,' starting index =',f19.0)
tm0 = second ()
call bcnrand (n, aa, x)
tm1 = second ()
write (6, 2) tm1 - tm0, np, (x(i), i = 1, np)
2 format ('CPU time =',f16.6/'first',i8,' elements of sequence ='/(4f19.15))
write (6, 3) np, (x(i), i = n - np + 1, n)
3 format ('final',i8,' elements of sequence ='/(4f19.15))
stop
end

! real*8 function second ()
! second = 0.d0
! return
! end

subroutine bcnrand (n, aa, x)

!   This routine generates a sequence of IEEE 64-bit floating-point pseudorandom
!   numbers in the range (0, 1), based on the recently discovered class of 
!   normal numbers described in the paper "Random Generators and Normal Numbers"
!   by DHB and Richard Crandall, available at http://www.nersc.gov/~dhbailey.

!   On certain systems with a fused multiply-add instruction, see !> comments.

!   David H. Bailey    2008-08-25

implicit none
integer i, ib, j, k, n
real*8 a, aa, d1, d2, d3, dd1(2), dd2(2), dd3(2), expm2, &
  p3i, r3i, t53, x(n)
parameter (p3i = 3.d0 ** 33, r3i = 1.d0 / p3i, t53 = 2.d0 ** 53)
external expm2

!   Check input parameters.

if (aa < p3i + 100 .or. aa > t53) then
  write (6, 1) aa
1 format ('bcnrand: error - A must satisfy 3^33 + 100 <= aa <= 2^53'/ &
  'value =',1p,d25.15)
  goto 200
endif

if (n > 2.d0 * p3i / 3.d0) then
  write (6, 2) n, 2.d0 * p3i / 3.d0
2 format ('bcnrand:  warning - number of elements exceeds period'/ &
  'n, period =',i12,f18.0)
endif

!   Calculate starting element.

d2 = expm2 (aa - p3i, p3i)
d3 = aint (0.5d0 * p3i)
call ddmuldd (d2, d3, dd1)
d1 = aint (r3i * dd1(1))
call ddmuldd (d1, p3i, dd2)
call ddsub (dd1, dd2, dd3)
d1 = dd3(1)
if (d1 < 0.d0) d1 = d1 + p3i

!   Calculate successive members of sequence.

do i = 1, n
  dd1(1) = t53 * d1
  dd1(2) = 0.d0
  d2 = aint (t53 * d1 / p3i)
  call ddmuldd (p3i, d2, dd2)
  call ddsub (dd1, dd2, dd3)
  d1 = dd3(1)
  if (d1 < 0.d0) d1 = d1 + p3i
  x(i) = r3i * d1
enddo

200 continue

return
end

function expm2 (p, am)

!   expm2 = 2^p mod am.  This routine uses a left-to-right binary scheme.

implicit none
integer i
real*8 am, amr, d1, d2, dd1(2), dd2(2), dd3(2), ddm(2), expm2, p, p1, p3i, &
  r3i, pt1, r, t53
parameter (p3i = 3.d0 ** 33, r3i = 1.d0 / p3i, t53 = 2.d0 ** 53)

amr = 1.d0 / am
p1 = p
pt1 = t53
r = 1.d0
ddm(1) = am
ddm(2) = 0.d0

110 continue

if (p1 .ge. pt1) then
! r = mod (2.d0 * r, am)
  call ddmuldd (2.d0, r, dd1)
  if (dd1(1) > am) then
    call ddsub (dd1, ddm, dd2)
    dd1(1) = dd2(1)
    dd1(2) = dd2(2)
  endif
  r = dd1(1)
  p1 = p1 - pt1
endif
pt1 = 0.5d0 * pt1
if (pt1 .ge. 1.d0) then
!  r = mod (r * r, am)
  call ddmuldd (r, r, dd1)
  dd2(1) = amr * dd1(1)
  d2 = aint (dd2(1))
  call ddmuldd (am, d2, dd2)
  call ddsub (dd1, dd2, dd3)
  r = dd3(1)
  if (r < 0.d0) r = r + am
  goto 110
endif

expm2 = r

return
end

subroutine ddmuldd (da, db, ddc)

!   This subroutine computes ddc = da x db.

implicit none
real*8 a1, a2, b1, b2, cona, conb, da, db, ddc(2), split, s1, s2
parameter (split = 134217729.d0)

!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.  If not sure, do the opposite.
!   On IBM systems, compile with -O3 -qstrict.

! s1 = da * db
! s2 = da * db - s1

!   This splits da and db into high-order and low-order words.

cona = da * split
conb = db * split
a1 = cona - (cona - da)
b1 = conb - (conb - db)
a2 = da - a1
b2 = db - b1
s1 = da * db
s2 = (((a1 * b1 - s1) + a1 * b2) + a2 * b1) + a2 * b2
!>
ddc(1) = s1
ddc(2) = s2

return
end subroutine

subroutine ddsub (dda, ddb, ddc)

!   This subroutine computes ddc = dda - ddb.

implicit none
real*8 dda(2), ddb(2), ddc(2)
real*8 e, t1, t2

!   Compute dda + ddb using Knuth's trick.

t1 = dda(1) - ddb(1)
e = t1 - dda(1)
t2 = ((-ddb(1) - e) + (dda(1) - (t1 - e))) + dda(2) - ddb(2)

!   The result is t1 + t2, after normalization.

ddc(1) = t1 + t2
ddc(2) = t2 - (ddc(1) - t1)
return
end subroutine

