program memtest

!   This tests memory by generating a large array using subroutine bcnrand,
!   regenerating the same sequence in a separate array, and then comparing the
!   two arrays.  This test can be repeated as often as desired.

!   The parameter m is the number of iterations of the test.  The parameter
!   maxerr is the number of errors after which execution is terminated.
!   The parameter n is the size of the two arrays, in 8-byte words.  The
!   default value of n below, namely 10,000,000, results in a program of size
!   approximately 160 Mbyte.

!   See comments below for complete details of using bcnrandx.
!   On certain systems with a fused multiply-add instruction, see !> comments.

!   David H Bailey    2008-08-25

!   Parameters:
!     m       Number of iterations of the test; default = 100.
!     maxerr  Number of errors after which run is terminated; default = 100.
!     n       Size of each of the two arrays, in 8-byte words; default =
!             10,000,000 (resulting in a program about 160 Mbytes long).

implicit none
integer i, j, kerr, m, maxerr, n
parameter (m = 100, maxerr = 100, n = 10000000)
real*8 a, aj, t1, t2, t3, x(n), y(n)
parameter (a = 3.d0**33 + 100.d0)
save

kerr = 0

do j = 1, m                                                                       
  write (6, *) 'start of trial ', j, m
  aj = (j-1)*m + a
  call bcnrand (n, aj, x)
  call bcnrand (n, aj, y)

  do i = 1, n
    if (x(i) /= y(i)) then
      write (6, '(i10,2z20,2i12)') i, x(i), y(i), loc(x(i)), loc(y(i))
      kerr = kerr + 1
      if (kerr == maxerr) stop
    endif
  enddo
enddo

stop
end

subroutine bcnrand (n, aa, x)

!   This routine generates a sequence of IEEE 64-bit floating-point pseudorandom
!   numbers in the range (0, 1), based on the recently discovered class of 
!   normal numbers described in the paper "Random Generators and Normal Numbers"
!   by DHB and Richard Crandall, available at http://www.nersc.gov/~dhbailey.

!   On certain systems with a fused multiply-add instruction, see !> comments.

!   David H. Bailey    2008-08-25

implicit none
integer i, ib, j, k, n
real*8 a, aa, d1, d2, d3, dd1(2), dd2(2), dd3(2), expm2, &
  p3i, r3i, t53, x(n)
parameter (p3i = 3.d0 ** 33, r3i = 1.d0 / p3i, t53 = 2.d0 ** 53)
external expm2

!   Check input parameters.

if (aa < p3i + 100 .or. aa > t53) then
  write (6, 1) aa
1 format ('bcnrand: error - A must satisfy 3^33 + 100 <= aa <= 2^53'/ &
  'value =',1p,d25.15)
  goto 200
endif

if (n > 2.d0 * p3i / 3.d0) then
  write (6, 2) n, 2.d0 * p3i / 3.d0
2 format ('bcnrand:  warning - number of elements exceeds period'/ &
  'n, period =',i12,f18.0)
endif

!   Calculate starting element.

d2 = expm2 (aa - p3i, p3i)
d3 = aint (0.5d0 * p3i)
call ddmuldd (d2, d3, dd1)
d1 = aint (r3i * dd1(1))
call ddmuldd (d1, p3i, dd2)
call ddsub (dd1, dd2, dd3)
d1 = dd3(1)
if (d1 < 0.d0) d1 = d1 + p3i

!   Calculate successive members of sequence.

do i = 1, n
  dd1(1) = t53 * d1
  dd1(2) = 0.d0
  d2 = aint (t53 * d1 / p3i)
  call ddmuldd (p3i, d2, dd2)
  call ddsub (dd1, dd2, dd3)
  d1 = dd3(1)
  if (d1 < 0.d0) d1 = d1 + p3i
  x(i) = r3i * d1
enddo

200 continue

return
end

function expm2 (p, am)

!   expm2 = 2^p mod am.  This routine uses a left-to-right binary scheme.

implicit none
integer i
real*8 am, amr, d1, d2, dd1(2), dd2(2), dd3(2), ddm(2), expm2, p, p1, p3i, &
  r3i, pt1, r, t53
parameter (p3i = 3.d0 ** 33, r3i = 1.d0 / p3i, t53 = 2.d0 ** 53)

amr = 1.d0 / am
p1 = p
pt1 = t53
r = 1.d0
ddm(1) = am
ddm(2) = 0.d0

110 continue

if (p1 .ge. pt1) then
! r = mod (2.d0 * r, am)
  call ddmuldd (2.d0, r, dd1)
  if (dd1(1) > am) then
    call ddsub (dd1, ddm, dd2)
    dd1(1) = dd2(1)
    dd1(2) = dd2(2)
  endif
  r = dd1(1)
  p1 = p1 - pt1
endif
pt1 = 0.5d0 * pt1
if (pt1 .ge. 1.d0) then
!  r = mod (r * r, am)
  call ddmuldd (r, r, dd1)
  dd2(1) = amr * dd1(1)
  d2 = aint (dd2(1))
  call ddmuldd (am, d2, dd2)
  call ddsub (dd1, dd2, dd3)
  r = dd3(1)
  if (r < 0.d0) r = r + am
  goto 110
endif

expm2 = r

return
end

subroutine ddmuldd (da, db, ddc)

!   This subroutine computes ddc = da x db.

implicit none
real*8 a1, a2, b1, b2, cona, conb, da, db, ddc(2), split, s1, s2
parameter (split = 134217729.d0)

!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.  If not sure, do the opposite.
!   On IBM systems, compile with -O3 -qstrict.

! s1 = da * db
! s2 = da * db - s1

!   This splits da and db into high-order and low-order words.

cona = da * split
conb = db * split
a1 = cona - (cona - da)
b1 = conb - (conb - db)
a2 = da - a1
b2 = db - b1
s1 = da * db
s2 = (((a1 * b1 - s1) + a1 * b2) + a2 * b1) + a2 * b2
!>
ddc(1) = s1
ddc(2) = s2

return
end subroutine

subroutine ddsub (dda, ddb, ddc)

!   This subroutine computes ddc = dda - ddb.

implicit none
real*8 dda(2), ddb(2), ddc(2)
real*8 e, t1, t2

!   Compute dda + ddb using Knuth's trick.

t1 = dda(1) - ddb(1)
e = t1 - dda(1)
t2 = ((-ddb(1) - e) + (dda(1) - (t1 - e))) + dda(2) - ddb(2)

!   The result is t1 + t2, after normalization.

ddc(1) = t1 + t2
ddc(2) = t2 - (ddc(1) - t1)
return
end subroutine

